import java.util.*;

/** 
 * Skeleton code for doubly-connected edge list data structure
 */ 
 
class DCEL {
 
  ArrayList<Vertex> vertices = new ArrayList<Vertex>();
  HashSet<HalfEdge> halfEdges = new HashSet<HalfEdge>();
  HashSet<Face> faces = new HashSet<Face>();
  
  Face outerFace = null;
  
  public DCEL() {
    outerFace = this.new Face();
  }
  
  Vertex newVertex(Point2d coordinates) { return this.new Vertex(coordinates); }
  HalfEdge newHalfEdge(Vertex origin, Face face) { return this.new HalfEdge(origin, face); }
  Face newFace() { return this.new Face(); }
  
  
  ArrayList<HalfEdge> getCycleFrom(HalfEdge e) {
    HalfEdge loopEdge = e;
    ArrayList<HalfEdge> retEdges = new ArrayList<HalfEdge>();
    do {
      retEdges.add(loopEdge);
      loopEdge = loopEdge.next;
    } while (loopEdge != e);
    return retEdges; 
  }
  
  /**
   * Splits the face incident to e1 and e2 by adding an edge between the origins of e1 and e2. 
   * If e1 and e2 are not incident to the same face, or e1 is the next or prev of e2, nothing should
   * happen (because the call isn't "legal"). 
   */ 
  void splitFace(HalfEdge e1, HalfEdge e2) {
    // 0. Check whether the call is legal and return if not. 
    if (e1.face != e2.face || e1.next == e2 || e1.prev == e2 || e1 == e2) {
      return;
    }
    // 1. Create a new face for the second half of the split. 
    Face newFace = this.newFace();
    
    // 2. Create new half-edges for the splitting edge:
    HalfEdge s1 = this.new HalfEdge(e2.origin, e1.face);
    HalfEdge s2 = this.new HalfEdge(e1.origin, newFace);
    s1.makeTwins(s2);
    
    // 3. Insert your new half-edges into the cyclic lists: 
    e2.prev.makeNext(s1);
    e1.prev.makeNext(s2);
    s1.makeNext(e1);
    s2.makeNext(e2);
    
    // 4. Make sure that all the half-edges point to the right face
    for (HalfEdge e : getCycleFrom(e2)) {
      e.face = newFace;
    }
    
    // 5. Make sure that each face correctly refers to an incident half-edge
    newFace.incidentEdge = e2;
    e1.face.incidentEdge = e1;
  }
  
  class HalfEdge {
    HalfEdge twin = null, next = null, prev = null; 
    Vertex origin = null; 
    Face face = null;
    
    private HalfEdge(Vertex origin, Face face) {
      this.origin = origin;
      this.face = face;
      halfEdges.add(this);
      origin.outgoingEdge = this;
      face.incidentEdge = this;
    }
    
    Vertex getDestination() {
      return twin.origin;
    }
    
    /**
     * Helper for making this the twin of that and that the twin of this. 
     */
    void makeTwins(HalfEdge that) {
      this.twin = that; 
      that.twin = this;
    }
    
    /** 
     * Helper for making that the next of this and this the prev of that. 
     */
    void makeNext(HalfEdge that) {
      this.next = that;
      that.prev = this;
    }
    
    void splitInHalf() {
      // 1. Create a vertex for the mid-point
      Vertex mid = new Vertex(midPoint(origin.coord, getDestination().coord));
      
      // 2. Create the two new half-edges
      HalfEdge n1 = new HalfEdge(mid, this.face);
      HalfEdge n2 = new HalfEdge(mid, this.twin.face);
      
      // 3. Set next/prev pointers
      n1.next = this.next; 
      n1.next.prev = n1;
      this.next = n1;
      n1.prev = this;
      
      n2.next = this.twin.next;
      n2.next.prev = n2;
      this.twin.next = n2;
      n2.prev = this.twin;
      
      // 4. Set twin pointers
      this.twin.twin = n1;
      n1.twin = this.twin;
      
      this.twin = n2;
      n2.twin = this;
    }
    
    // SOLUTION
    
    boolean isLegal() {
      if (this.face == outerFace || this.twin.face == outerFace) return true;
      Point2d a = this.origin.coord; 
      Point2d c = this.next.origin.coord;
      Point2d d = this.prev.origin.coord;
      Point2d b = this.twin.next.next.origin.coord;
      return !inCircle(a,b,c,d);
    }
    
    void remove() {
      if (this.face == outerFace || 
      this.twin.face == outerFace || 
      this.face == this.twin.face) return;
      
      DCEL.HalfEdge next1 = this.next;
      DCEL.HalfEdge prev1 = this.twin.prev;
      DCEL.HalfEdge next2 = this.twin.next;
      DCEL.HalfEdge prev2 = this.prev;
      
      prev1.makeNext(next1);
      prev2.makeNext(next2);
      
      DCEL.this.halfEdges.remove(this);
      DCEL.this.halfEdges.remove(this.twin);
      
      DCEL.this.faces.remove(next2.face);
      for (HalfEdge e : getCycleFrom(next1)) {
        e.face = next1.face;
      }
      next1.face.incidentEdge = next1;
      
    }
    
    void flip() {
      
      if (this.face == DCEL.this.outerFace || this.twin.face == DCEL.this.outerFace) {
        return;
      }
      
      Point2d a = this.origin.coord; 
      Point2d c = this.next.origin.coord;
      Point2d d = this.prev.origin.coord;
      Point2d b = this.twin.next.next.origin.coord;
      
      if (isRightHandTurn(a, b, c) ||
          isRightHandTurn(b, c, d) ||
          isRightHandTurn(c, d, a) ||
          isRightHandTurn(d, a, b)) return;
      
      HalfEdge A = this.next.next;
      HalfEdge B = this.twin.next.next;
      remove();
      splitFace(A, B);
    }
    
    // END SOLUTION
    
  }
  
  class Vertex {
    Point2d coord;
    HalfEdge outgoingEdge; 
    private  Vertex(Point2d coordinates) {
      vertices.add(this);
      this.coord = coordinates; 
    }
  }
  
  class Face {
    HalfEdge incidentEdge; 
    private Face() {
      faces.add(this);
    }
  }
  
}